#include "defines.h"
#include "bits.h"
#include "clunet.h"
#include "usart.h"
#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>

struct packet_t 
{
	char busy;
	unsigned char src_address;
	unsigned char dst_address;
	unsigned char command;
	unsigned char size;
	char data[CLUNET_READ_BUFFER_SIZE];
};

volatile struct packet_t read_buffer[PACKETS_BUFFER];
volatile char write_buffer[CLUNET_SEND_BUFFER_SIZE];
volatile int usart_read_pos = 0;
volatile char crc_error = 0;
volatile char sending = 0;

static inline char
check_crc(const char* data, const uint8_t size)
{
      uint8_t crc = 0;
      uint8_t i, j;
      for (i = 0; i < size; i++)
      {
            uint8_t inbyte = data[i];
            for (j = 0 ; j < 8 ; j++)
            {
                  uint8_t mix = (crc ^ inbyte) & 1;
                  crc >>= 1;
                  if (mix) crc ^= 0x8C;
                  inbyte >>= 1;
            }
      }
      return crc;
}

ISR(USART_RXC_vect)
{
	unsigned char b, c = 0;
	while (UCSRA & (1<<RXC))
	{
		b = UDR;
//		USART_TransmitByte(b);
		if (b >= '0' && b <= '9') c = b-'0';
		else if (b >= 'A' && b <= 'F') c = b-'A'+10;
		else if (b == '\r' || b == '\n')
		{	
			if (usart_read_pos > 0) //    
			{
				if ((usart_read_pos >= 8) &&
					(check_crc((char*)write_buffer, usart_read_pos/2) == 0))
				{
					int status = clunet_ready_to_send();
					if (status == 0 || status < write_buffer[0])
					{
						clunet_send(write_buffer[1], write_buffer[0], write_buffer[2], (char*)(write_buffer+3), usart_read_pos/2-4);
						sending = 1;
						LED_WRITE_ON;
					}
				
				} else {
					crc_error = sending = 1;
				}
			}
			usart_read_pos = 0;
			continue;
		}
		else continue;
		if (usart_read_pos/2 < CLUNET_SEND_BUFFER_SIZE)
		{
			if (usart_read_pos % 2 == 0) write_buffer[usart_read_pos/2] = (c<<4);
				else write_buffer[usart_read_pos/2] |= (c & 0x0F);
			usart_read_pos++;
		}
	}
}

void data_received(unsigned char src_address, unsigned char dst_address, unsigned char command, char* data, unsigned char size)
{		
	unsigned char i, m;
	for (i = 0; i < PACKETS_BUFFER; i++)
	if (!read_buffer[i].busy)
	{
		read_buffer[i].busy = 1;
		read_buffer[i].src_address = src_address;
		read_buffer[i].dst_address = dst_address;
		read_buffer[i].command = command;
		read_buffer[i].size = size;
		for (m = 0; m <= size; m++) read_buffer[i].data[m] = data[m];
		return;
	}
}

int main (void)
{
	int i, m;
	for (i = 0; i < PACKETS_BUFFER; i++) read_buffer[i].busy = 0;
	set_bit(DDRB,0); set_bit(DDRB,1);
	USART_init();
	clunet_init();
	clunet_set_on_data_received_sniff(&data_received);
	LED_WRITE_ON; LED_READ_ON;
	_delay_ms(500);
	LED_WRITE_OFF; LED_READ_OFF;
	sei();
	while(1)
	{	
		for (i = 0; i < PACKETS_BUFFER; i++)
		if (read_buffer[i].busy)
		{
			if (read_buffer[i].src_address != CLUNET_DEVICE_ID) LED_READ_ON;
			USART_TransmitHex(read_buffer[i].src_address); USART_TransmitByte(' ');				
			USART_TransmitHex(read_buffer[i].dst_address); USART_TransmitByte(' ');
			USART_TransmitHex(read_buffer[i].command); USART_TransmitByte(' ');
			USART_TransmitHex(read_buffer[i].size); USART_TransmitByte(' ');
			for (m = 0; m < read_buffer[i].size; m++)
				USART_TransmitHex(read_buffer[i].data[m]);
			USART_TransmitByte(' ');
			USART_TransmitHex(read_buffer[i].data[read_buffer[i].size]);
			USART_TransmitByte('\r');
			USART_TransmitByte('\n');
			read_buffer[i].busy = 0;
			LED_READ_OFF;
		}
		if (crc_error)
		{
			USART_TransmitText("!CRC ERROR\r\n");
			crc_error = 0;
		}

		if (sending)
		{
			if (clunet_ready_to_send() == 0)
			{
				//USART_TransmitText("!\r\n");
				sending = 0;
				LED_WRITE_OFF;				
			}
		}				
	}
	return 0;
}
